# tkGeoMapProcs - provide some general procedures and bindings to use with
# the tkGeoMap extension to wish.
# 
# Procedures defined in this file:
#
# Projections:
# proc Mercator wdg
#   Set projection for geomap wdg to Mercator.  Mouse drag moves map.
# proc LambertConfConic wdg
#   Set projection for geomap wdg to Lambert Conformal Conic.
#   Double-1 moves map.
# proc Sinusoidal wdg
#   Set projection for geomap wdg to Sinusoidal.
#   Double-1 moves map.
# proc Orthographic wdg
#   Set projection for geomap wdg to Orthographic.
#   Double-1 moves map.
# proc PolarStereographicN wdg
#   Set projection for geomap wdg to Polar Stereographic, north.
#   Double-1 switches to LambertConformal Conic.
# proc PolarStereographicS wdg
#   Set projection for geomap wdg to Polar Stereographic, south.
#   Double-1 switches to LambertConformal Conic.
#
# Print utility
# proc PrintGeoMap wdg
#   X Window dump a geomap widget window
#
# GeoMap in a canvas with a menu
# proc MapCanvas fm userParm
#   Create a canvas in frame fm.  Create a geomap and add it as a window
#   to the canvas.  Create a menu in the frame to manipulate the geomap.
#   Array userParm contains initial values for the geomap, otherwise, defaults
#   are used.  See documentation at start of MapCanvas procedure definition
#   for accepted values of userParm.
#
# Following procs are mainly used by MapCanvas, although they may have general
# utility.
# proc ShowMapInfo map
#   Create a string with information about geomap named map.  Used in
#   MapCanvas.
# proc PrintCanvasMap {canvas map wMargin nMargin
#   X Window dump a MapCanvas
# proc Circle {center radius {unit gsdeg}}
#   Generate a series of points at given radius from center (a lat-lon)
#   Can give distance unit for radius, otherwise radius assumed to be great
#   circle degrees.
# proc EraseCircle wdg
#   Erase geoLineArray namesd circle$wdg, if it exists.

package provide tkGeoMapProcs 1.2

proc Mercator wdg {
  # Set Mercator projection with bindings
  $wdg Projection Mercator

  EraseCircle $wdg

  # Remember button press location
  bind $wdg <Button-1> {
    set x0 %x
    set y0 %y
    set ctr0 [%W xtolatlon "%x %y"]
    set pctr0 [%W latlontoproj [%W center]]
    set ord0 [lindex $pctr0 0]
    set abs0 [lindex $pctr0 1]
  }
    
  # Move map when button-1 moves
  bind $wdg <B1-Motion> {
    set ord [expr $ord0 - ($y0 - %y) * $CmPerPx]
    set abs [expr $abs0 - (%x - $x0) * $CmPerPx]
    %W center [%W projtolatlon "$ord $abs"]
  }

  # Adjust projection at button-1 release
  bind $wdg <ButtonRelease-1> {
    set ord [expr [%W cget -height] / 2]
    set abs [expr [%W cget -width] / 2]
    set ctr [%W xtolatlon "$abs $ord"]
    set rlon [lindex $ctr 1]
    %W projection Mercator $rlon
  }

  bind $wdg <Double-1> {}
  puts "Drag to move map"
}

proc LambertConfConic wdg {
  # Set Lambert Conformal Conic with bindings
  $wdg Projection LambertConfConic [$wdg center]

  EraseCircle $wdg

  bind $wdg <Button-1> { }
  bind $wdg <ButtonRelease-1> { }
  bind $wdg <B1-Motion> { }

  # Double-1 chooses location for new map center
  bind $wdg <Double-1> {
    set refPoint [%W XToLatLon "%x %y"]
    %W center $refPoint
    %W Projection LambertConfConic $refPoint
  }
  puts "Double-1 to move map"
}

proc Sinusoidal wdg {
  # Set Sinusoidal with bindings
  $wdg Projection Sinusoidal [lindex [$wdg center] 1]

  EraseCircle $wdg

  bind $wdg <Button-1> { }
  bind $wdg <ButtonRelease-1> { }
  bind $wdg <B1-Motion> { }

  # Double-1 chooses location for new map center
  bind $wdg <Double-1> {
    set refPoint [%W XToLatLon "%x %y"]
    set rlon [lindex $refPoint 1]
    %W center $refPoint
    %W Projection Sinusoidal $rlon
  }
  puts "Double-1 to move map"
}

proc Orthographic wdg {
  # Set Orthographic with bindings
  set center [$wdg center]
  $wdg Projection Orthographic $center

  bind $wdg <Button-1> { }
  bind $wdg <ButtonRelease-1> { }
  bind $wdg <B1-Motion> { }

  # Draw circle around the map
  EraseCircle $wdg
  set circleName circle$wdg
  geolinearray fmlist $circleName [Circle $center 89.9]
  $wdg lines $circleName

  # Double-1 chooses location for new map center
  bind $wdg <Double-1> {
    set refPoint [%W XToLatLon "%x %y"]
    set rlat [lindex $refPoint 0]
    set rlon [lindex $refPoint 1]
    if {abs($rlat) <= 90.0 && abs($rlon) <= 360.0} {
      # Encircle the world
      set circleName circle%W
      %W eraselines $circleName
      geolinearray rm $circleName
      geolinearray fmlist $circleName [Circle $refPoint 89.9]
      %W lines $circleName

      # Draw the world
      %W center $refPoint
      %W Projection Orthographic $refPoint
    } else {
      puts "Selected point not in domain"
    }
  }
  puts "Double-1 to move map"
}

proc PolarStereographicN wdg {
  # Set Polar Stereographic at North Pole with bindings
  $wdg Projection PolarStereographic N

  EraseCircle $wdg

  bind $wdg <Button-1> { }
  bind $wdg <ButtonRelease-1> { }
  bind $wdg <B1-Motion> { }

  # Double-1 set Lambert Conformal Conic
  bind $wdg <Double-1> {
    set refPoint [%W XToLatLon "%x %y"]
    %W center $refPoint
    LambertConfConic %W
  }
  puts "Double-1 to move map"
}

proc PolarStereographicS wdg {
  # Set Polar Stereographic at North Pole with bindings
  $wdg Projection PolarStereographic S

  EraseCircle $wdg

  bind $wdg <Button-1> { }
  bind $wdg <ButtonRelease-1> { }
  bind $wdg <B1-Motion> { }

  # Double-1 set Lambert Conformal Conic
  bind $wdg <Double-1> {
    set refPoint [%W XToLatLon "%x %y"]
    %W center $refPoint
    LambertConfConic %W
  }
  puts "Double-1 to move map"
}

proc PrintGeoMap wdg {
  # X Window dump a geomap widget window
  set thick 2

  set  winWidth [expr [$wdg cget  -width] - 1]
  set winHeight [expr [$wdg cget -height] - 1]

  set topLeft [$wdg XToLatLon "1 1"]
  set topRight [$wdg XToLatLon "$winWidth 1"]
  set bottomRight [$wdg XToLatLon "$winWidth $winHeight"]
  set bottomLeft [$wdg XToLatLon "1 $winHeight"]
  set border [list $topLeft $topRight $bottomRight $bottomLeft $topLeft]

  geolinearray fmlist border $border
  $wdg lines border -h $thick

  set bd [$wdg cget -bd]
  $wdg configure -bd 0
  $wdg update now

  exec xwd -id [winfo id $wdg] > geomap.d
    
  $wdg eraselines border
  geolinearray rm border
  $wdg configure -bd $bd
}

proc MapCanvas {fm userParm} {
  # Create a frame called $fm containing a canvas widget called $fm.c with a 
  # geomap called $fm.c.g embedded in it and a menu for manipulating the 
  # geomap widget
  #
  # userParm is the name of an array which can contain optional user
  # selected # parameters.  It is upvar'ed to local variable params.
  # If element is not in array, proc uses a default value
  # The following parameter values are recognized.
  #   params(cvsWidth)    = canvas width
  #   params(cvsHeight)   = canvas height
  #   params(center)      = center
  #   params(scale)       = initial scale
  #   params(projection)  = initial projection
  #   params(mapMgn)      = blank space in the canvas around the embedded map
  #                         4-element list {top right bottom left}
  #   params(projections) = list of projections to put in projections menu,
  #                         or "none" if projections menu not wanted.
  #                         If not set, then no projections menu created.
  #   params(scales)      = list of scales to put in the scales menu,
  #                         or "none" if projections menu not wanted.
  #                         If not set, then no scales menu created.
  #   params(dotSizes)    = list of integers in dot size menu, or "none"
  #                         if no dotsize menu wanted.
  #   params(rotateMenu)  = Boolean switch.  Non-zero => use "Top" menu

  upvar $userParm params
  if [array exists params] {
    set paramList [array names params]
  } else {
    set paramList "none"
  }
  
  # Initial size of canvas containing the geomap widget
  set cvsWidth [expr {[lsearch $paramList cvsWidth] >= 0 \
                      ? $params(cvsWidth) : 400}]
  set cvsHeight [expr {[lsearch $paramList cvsHeight] >= 0 \
                      ? $params(cvsHeight) : 400}]

  # Initial center
  set center [expr {[lsearch $paramList center] >= 0 \
                     ? $params(center) : "0.000 0.000"}]

  # Initial scale
  set mapScale [expr {[lsearch $paramList scale] >= 0 \
                     ? $params(scale) : "1:30000000"}]

  # Initial projection
  set projection [expr {[lsearch $paramList projection] >= 0 \
                       ? $params(projection) : "Mercator"}]
  
  # mapMgn: Extra space in the canvas around the map
  #   Is a list: {top right bottom left}.  Make global for use in bind commands.
  global mapMgn
  if {[lsearch $paramList mapMgn] >= 0} {
    set mapMgn(${fm}.c) $params(mapMgn)
  } else {
    set mapMgn(${fm}.c) {50 10 10 10}
  }

  # Projection menu
  if {[lsearch $paramList projections] == -1} {
    # If projections not in params, use defaults
    set projections {Mercator LambertConfConic PolarStereographicN \
      PolarStereographicS Sinusoidal Orthographic}
  } elseif {$params(projections) != "none"} {
    # User provided projection list
    set projections $params(projections)
  }

  # Scale menu
  if {[lsearch $paramList scales] == -1} {
    # If scales not in params, use defaults
    set scales {1:10000000 1:20000000 1:30000000 1:60000000 1:90000000 \
      1:120000000 1:240000000}
  } elseif {$params(scales) != "none"} {
    # User provided scale selections
    set scales $params(scales)
  }

  # Dot size menu
  if {[lsearch $paramList dotSizes] == -1} {
    set dotSizes {0 1 2 3}
  } elseif {$params(dotSizes) != "none"} {
    set dotSizes $params(dotSizes)
  }

  # Rotation menu
  if {[lsearch $paramList rotateMenu] == -1} {
    set rotateMenu 1
  } else {
    set rotateMenu $params(rotateMenu)
  }

  ############ End of parameters ############
  
  # Create a canvas
  canvas $fm.c -width  $cvsWidth -height $cvsHeight

  # Create a geomap
  set topMgn     [lindex $mapMgn(${fm}.c) 0]
  set rightMgn   [lindex $mapMgn(${fm}.c) 1]
  set bottomMgn  [lindex $mapMgn(${fm}.c) 2]
  set leftMgn    [lindex $mapMgn(${fm}.c) 3]
  geomap $fm.c.g \
    -width [expr $cvsWidth - $leftMgn - $rightMgn] \
    -height [expr $cvsHeight - $topMgn - $bottomMgn]

  # Set center, projection and scale
  $fm.c.g center $center
  $projection $fm.c.g
  $fm.c.g configure -scale $mapScale
  
  # Put the map into the canvas
  $fm.c create window $leftMgn $topMgn -window $fm.c.g -anchor nw
  $fm.c create text $leftMgn 5 -anchor nw -tags mapInfo
  
  $fm.c itemconfigure mapInfo -text [ShowMapInfo $fm.c.g]
  
  # Handle configure events
  bind $fm.c <Configure> {
    set cvsWidth [winfo width %W]
    set cvsHeight [winfo height %W]
    set topMgn     [lindex $mapMgn(%W) 0]
    set rightMgn   [lindex $mapMgn(%W) 1]
    set bottomMgn  [lindex $mapMgn(%W) 2]
    set leftMgn    [lindex $mapMgn(%W) 3]
    %W.g configure -width [expr $cvsWidth - $leftMgn - $rightMgn]
    %W.g configure -height [expr $cvsHeight - $topMgn - $bottomMgn]
  }
  
  # Create menu bar and menus
  frame $fm.mbar -relief raised -bd 2
  
  # Widget menu
  menubutton $fm.mbar.wdg -text Widget -menu $fm.mbar.wdg.menu
  menu $fm.mbar.wdg.menu
  $fm.mbar.wdg.menu add command -label print \
  	-command "PrintCanvasMap $fm.c $fm.c.g $leftMgn $topMgn"
  $fm.mbar.wdg.menu add command -label exit -command exit
  set menuList $fm.mbar.wdg
  
  # Projection menu
  if [info exists projections] {
    menubutton $fm.mbar.projection -text Projection \
      -menu $fm.mbar.projection.menu
    menu $fm.mbar.projection.menu
    foreach projection $projections {
      $fm.mbar.projection.menu add command -label $projection \
        -command "$projection $fm.c.g
                  $fm.c itemconfigure mapInfo -text \[ShowMapInfo $fm.c.g\]"
    }
    lappend menuList $fm.mbar.projection
  }
  
  # Map scale menu
  if [info exists scales] {
    menubutton $fm.mbar.scale -text Scale -menu $fm.mbar.scale.menu
    menu $fm.mbar.scale.menu
    foreach scale $scales {
      $fm.mbar.scale.menu add command -label $scale \
        -command "$fm.c.g configure -scale $scale
                  $fm.c itemconfigure mapInfo -text \[ShowMapInfo $fm.c.g\]"
    }
    lappend menuList $fm.mbar.scale
  }
  
  # Dot size menu
  if [info exists dotSizes] {
    menubutton $fm.mbar.dotsize -text DotSize -menu $fm.mbar.dotsize.menu
    menu $fm.mbar.dotsize.menu
    foreach dotSize $dotSizes {
      $fm.mbar.dotsize.menu add command -label "  $dotSize" \
        -command "$fm.c.g configure -dotsize $dotSize"
    }
    lappend menuList $fm.mbar.dotsize
  }
    
  # Rotation angle menu
  # User selects which compass direction is "up."
  if {$rotateMenu} {
    menubutton $fm.mbar.rotation -text Top -menu $fm.mbar.rotation.menu
    menu $fm.mbar.rotation.menu
    set bearing(\ n)    0
    set bearing(ne)   -45
    set bearing(\ e)  -90
    set bearing(se)  -135
    set bearing(\ s)  180
    set bearing(sw)   135
    set bearing(\ w)   90
    set bearing(nw)    45
    foreach rotation {{ n} {ne} { e} {se} { s} {sw} { w} {nw}} {
      $fm.mbar.rotation.menu add command -label "$rotation" \
        -command "$fm.c.g rotation $bearing($rotation)
                  $fm.c itemconfigure mapInfo -text \[ShowMapInfo $fm.c.g\]"
    }
    lappend menuList $fm.mbar.rotation
  }

  # Pack the widgets
  pack $fm.mbar -side top -fill x
  eval pack $menuList -side left
  pack $fm.c -expand true -fill both
}

proc ShowMapInfo map {
  # Make a string with map information
  set projInfo [lindex [$map projection] 0]
  set scaleInfo "Scale [$map cget -scale]."
  set rotation [lindex [$map rotation] 0]
  if {$rotation == 0.0} {
    set rotationInfo ""
  } else {
    set rotation \
      [expr $rotation - 360.0 * ceil(($rotation - 180.0) / 360.0)]
    set rotationInfo "Rotated $rotation degrees."
  }
  return "$projInfo\n$scaleInfo  $rotationInfo"
}

# X Window Dump a canvas and its embedded geomap
proc PrintCanvasMap {canvas map wMargin nMargin} {

  set width [$map cget -width]
  set height [$map cget -height]

  # Replace shaded border with drawn rectangle
  set bd [$map cget -bd]
  $map configure -bd 0
  set border [$canvas create rectangle $wMargin $nMargin \
    [expr $wMargin + $width] [expr $nMargin + $height] -width 2]
  update

  # Call X window dump and X print
  exec xwd -id [winfo id $canvas] > mapcanvas.d

  # Restore shaded border
  $canvas delete $border
  $map configure -bd $bd
}

proc Circle {center radius {unit gsdeg}} {
  # Return list of points at distance radius from center
  for {set az 0} {$az <= 360} {incr az 3} {
    lappend circle [geoplace step $center $az $radius $unit]
  }
  return $circle
}

proc EraseCircle wdg {
  set circleName circle$wdg
  if {[lsearch [geolinearray list] $circleName] >= 0} {
    $wdg eraselines $circleName
    geolinearray rm $circleName
  }
}
